import abc
import collections
import datetime
import copy
import itertools
import logging
import re

import urllib

from .exceptions import ClientValidationError
from .utils import make_date, is_bound


logger = logging.getLogger(__name__)


class BaseManager(metaclass=abc.ABCMeta):
    """
    Base class for Model Manager classes.

    Instantiated by the Model metaclass and attached to model._meta.manager.
    The Client object needs to set itself on the Manager instance before it's used.
    """

    # _URL_KEY = None
    model = None

    def __init__(self, client):
        self.client = client

    def _meta_attribute(self, attribute, default=None):
        return getattr(self.model._meta, attribute, default)

    def create_from_result(self, result, **kwargs):
        obj = self.model()
        return obj._deserialize(result, self, **kwargs)

    def _get(self, target_url, expand=[]):
        headers = {}
        if expand:
            headers["Expand"] = ",".join(expand)

        r = self.client.request("GET", target_url, headers=headers)
        return self.create_from_result(r.json())

    def _reverse_url(self, url):
        return self.client.reverse_url(self._URL_KEY, url)


class InnerManager(BaseManager, metaclass=abc.ABCMeta):
    def __init__(self, client, parent_manager):
        super(InnerManager, self).__init__(client)
        self.parent = parent_manager


class Manager(BaseManager, metaclass=abc.ABCMeta):
    def list(self):
        """
        Fetches a set of model objects

        :rtype: :py:class:`koordinates.base.Query`
        """
        target_url = self.client.get_url(self._URL_KEY, "GET", "multi")
        return Query(self, target_url)

    def get(self, id, expand=[]):
        """Fetches a Model instance determined by the value of `id`.

        :param id: numeric ID for the Model.
        """
        target_url = self.client.get_url(self._URL_KEY, "GET", "single", {"id": id})
        return self._get(target_url, expand=expand)

    # Query methods we delegate
    def filter(self, *args, **kwargs):
        """
        Returns a filtered Query view of the model objects.
        Equivalent to calling ``.list().filter(...)``.
        See :py:meth:`koordinates.base.Query.filter`.
        """
        return self.list().filter(*args, **kwargs)

    def order_by(self, *args, **kwargs):
        """
        Returns an ordered Query view of the model objects.
        Equivalent to calling ``.list().order_by(...)``.
        See :py:meth:`koordinates.base.Query.order_by`.
        """
        return self.list().order_by(*args, **kwargs)

    def expand(self, *args, **kwargs):
        """
        Returns an expanded Query view of the model objects.
        Equivalent to calling ``.list().expand()``.
        Using expansions may have significant performance implications for some API requests.
        See :py:meth:`koordinates.base.Query.expand`.
        """
        return self.list().expand(*args, **kwargs)

    def create(self, object):
        # We don't have to have creatable models, so it's not an abstractmethod.
        raise NotImplementedError()


class Query(object):
    """
    Chainable query class to manage GET requests for lists.
    Supports filtering by multiple attributes and sorting, and deals
    with pagination of resultsets into a single iterator.

    Query objects are instantiated by methods on the Manager classes.
    To actually execute the query, iterate over it. You can also call
    len() to return the length - this will do the "first" page request
    and examine the ``X-Resource-Range`` header to produce a count.
    """

    def __init__(
        self, manager, url, valid_filter_attributes=None, valid_sort_attributes=None
    ):
        self._manager = manager
        self._target_url = url
        self._count = None
        self._filters = collections.defaultdict(list)
        self._order_by = None
        self._expand = None
        self._extra = collections.defaultdict(list)

        self._valid_filter_attrs = (
            self._manager._meta_attribute("filter_attributes", [])
            if valid_filter_attributes is None
            else valid_filter_attributes
        )
        self._valid_sort_attrs = (
            self._manager._meta_attribute("ordering_attributes", [])
            if valid_sort_attributes is None
            else valid_sort_attributes
        )

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self._manager.model.__name__)

    def __str__(self):
        return self._to_url()

    def _request(self, url, method="GET"):
        r = self._manager.client.request(method, url, headers=self._to_headers())
        r.raise_for_status()
        return r

    def _update_range(self, response):
        """ Update the query count property from the `X-Resource-Range` response header """
        header_value = response.headers.get("x-resource-range", "")
        m = re.match(r"\d+-\d+/(\d+)$", header_value)
        if m:
            self._count = int(m.group(1))
        else:
            self._count = None

    def _to_url(self):
        """ Serialises this query into a request-able URL including parameters """
        url = self._target_url

        params = collections.defaultdict(list, copy.deepcopy(self._filters))
        if self._order_by is not None:
            params["sort"] = self._order_by
        for k, vl in list(self._extra.items()):
            params[k] += vl

        if params:
            url += "?" + urllib.parse.urlencode(params, doseq=True)

        return url

    def _to_headers(self):
        """ Serialises this query into a request-able set of headers """
        headers = {}
        if self._expand:
            headers["Expand"] = "list"

        return headers

    def _next_url(self, response):
        """
        Return the URL to the next page of results

        Paginate via Link headers
        Link URLs will include the query parameters, so we can use it as an entire URL.
        """
        return response.links.get("page-next", {}).get("url", None)

    def __iter__(self):
        """
        Execute this query and return the results (generally as Model objects)
        """
        if hasattr(self, "_first_page"):
            # if len() has been called on this Query, we have a cached page
            # of results & a next url
            page_results, url = self._first_page
            del self._first_page
        else:
            url = self._to_url()
            r = self._request(url)
            page_results = r.json()

            # Update position
            self._update_range(r)

            # Point to the next page
            url = self._next_url(r)

        for raw_result in page_results:
            yield self._manager.create_from_result(raw_result)

        while url:
            r = self._request(url)
            page_results = r.json()

            # Update position
            self._update_range(r)

            for raw_result in page_results:
                yield self._manager.create_from_result(raw_result)

            # Paginate via Link headers
            # Link URLs will include the query parameters, so we can use it as an entire URL.
            url = r.links.get("page-next", {}).get("url", None)

    def __len__(self):
        """
        Get the count for the query results. If we've previously started iterating we use
        that count, otherwise do a request and look at the ``X-Resource-Range`` header.
        """
        # note: this method is called twice when wrapping queries in list(), from py3.8+
        # https://bugs.python.org/issue39829
        # so we need to make sure it is cached and doesn't do a request every time.
        if self._count is None:
            r = self._request(self._to_url())
            self._update_range(r)
            self._first_page = (r.json(), self._next_url(r))
            if self._count is None and self._first_page[1] is None:
                # this is the only page
                self._count = len(self._first_page[0])
        return self._count

    def __getitem__(self, k):
        """
        Limited slicing support ([N] and [:N] only, for positive N)
        """
        if isinstance(k, int) and k >= 0:
            try:
                return next(itertools.islice(self.__iter__(), k, None))
            except StopIteration:
                raise IndexError(k)

        elif (
            not isinstance(k, slice)
            or k.start is not None
            or k.step is not None
            or k.stop is None
            or k.stop <= 0
        ):
            raise ValueError("Only query[:+N] or query[+N] slicing is supported.")

        return list(itertools.islice(self.__iter__(), k.stop))

    def _clone(self):
        q = Query(
            manager=self._manager,
            url=self._target_url,
            valid_filter_attributes=self._valid_filter_attrs,
            valid_sort_attributes=self._valid_sort_attrs,
        )

        q._filters = collections.defaultdict(list, copy.deepcopy(self._filters))
        q._order_by = self._order_by
        q._expand = self._expand
        q._extra = collections.defaultdict(list, copy.deepcopy(self._extra))
        return q

    def extra(self, **params):
        """
        Set extra query parameters (eg. filter expressions/attributes that don't validate).
        Appends to any previous extras set.

        :rtype: Query
        """
        q = self._clone()
        for key, value in list(params.items()):
            q._extra[key].append(value)
        return q

    def filter(self, **filters):
        """
        Add a filter to this query.
        Appends to any previous filters set.

        :rtype: Query
        """

        q = self._clone()
        for key, value in list(filters.items()):
            filter_key = re.split("__", key)
            filter_attr = filter_key[0]
            if filter_attr not in self._valid_filter_attrs:
                raise ClientValidationError("Invalid filter attribute: %s" % key)

            # we use __ as a separator in the Python library, the APIs use '.'
            q._filters[".".join(filter_key)].append(value)
        return q

    def order_by(self, sort_key=None):
        """
        Set the sort for this query. Not all attributes are sorting candidates.
        To sort in descending order, call ``Query.order_by('-attribute')``.

        Calling ``Query.order_by()`` replaces any previous ordering.

        :rtype: Query
        """
        if sort_key is not None:
            sort_attr = re.match(r"(-)?(.*)$", sort_key).group(2)
            if sort_attr not in self._valid_sort_attrs:
                raise ClientValidationError("Invalid ordering attribute: %s" % sort_key)

        q = self._clone()
        q._order_by = sort_key
        return q

    def expand(self):
        """
        Expand list results in this query.
        This can have a performance penalty for some objects.

        :rtype: Query
        """
        q = self._clone()
        q._expand = True
        return q


class ModelMeta(type):
    """
    Sets up the special model characteristics based on the ``Meta:`` object on the model
    """

    def __new__(meta, name, bases, attrs):
        klass = super(ModelMeta, meta).__new__(meta, name, bases, attrs)
        try:
            ModelBase
            Model
            InnerModel
        except NameError:
            # klass is ModelBase/Model/InnerModel, it doesn't have a `Meta:` object
            pass
        else:
            if not hasattr(klass, "Meta"):
                raise AttributeError("Missing Meta object for %s" % klass)

            # Move model.Meta to model._meta
            klass._meta = klass.Meta
            del klass.Meta

            # Associate this model with it's manager
            if (
                getattr(klass._meta.manager, "model")
                and klass._meta.manager.model is not klass
            ):
                # the manager already has a model!
                if issubclass(klass, klass._meta.manager.model):
                    # this is due to subclassing the model and not the manager
                    # we *don't* add the subclass, we assume you're doing downcasting or
                    # some other cleverness with the subclass, and only querying the superclass.
                    pass
                else:
                    raise TypeError(
                        "%s already has an associated model: %s"
                        % (
                            klass._meta.manager.__name__,
                            klass._meta.manager.model.__name__,
                        )
                    )
            else:
                klass._meta.manager.model = klass

            # add default accessors for relations
            # class MyModel(Model):
            #     class Meta:
            #         relations = {
            #             'foo': FooModel,    # FK to a single FooModel
            #             'bars': [BarModel], # pointer to multiple BarModel objects
            #         }
            #
            # -> mymodel.get_foo()
            # -> mymodel.list_bars()

            # method factories to workaround late-binding in loops
            def build_multi_getter(r_class, r_attr):
                @is_bound
                def _getter(self):
                    rel_url = getattr(self, r_attr)
                    rel_mgr = self._manager.client.get_manager(r_class)
                    return Query(rel_mgr, rel_url)

                return _getter

            def build_single_getter(r_class, r_attr):
                @is_bound
                def _getter(self):
                    rel_url = getattr(self, r_attr, None)
                    if rel_url:
                        rel_mgr = self._manager.client.get_manager(r_class)
                        return rel_mgr._get(rel_url)
                    else:
                        return None

                return _getter

            for ref_attr, ref_class in list(
                getattr(klass._meta, "relations", {}).items()
            ):
                if isinstance(ref_class, (list, tuple)) and len(ref_class) == 1:
                    # multiple relation
                    ref_method = "list_%s" % ref_attr
                    ref_class = ref_class[0]
                    ref_getter = build_multi_getter(ref_class, ref_attr)
                else:
                    # single relation
                    ref_method = "get_%s" % ref_attr
                    ref_getter = build_single_getter(ref_class, ref_attr)

                    setattr(klass, ref_method, ref_getter)

                # don't redefine any existing methods
                if not hasattr(klass, ref_method):
                    setattr(klass, ref_method, ref_getter)
                    logger.debug(
                        "klass=%s added relation method %s()", klass, ref_method
                    )
        return klass


class SerializableBase(object):
    """
    Base class for simple serialization.
    """

    def __setattr__(self, name, value):
        if isinstance(value, ModelBase) and not name.startswith("_"):
            # set the ._parent attribute on the passed-in Model instance
            object.__setattr__(value, "_parent", self)
        object.__setattr__(self, name, value)

    def _deserialize(self, data):
        """
        Deserialise from JSON response data.

        String items named ``*_at`` are turned into dates.

        Filters out:
        * attribute names in ``Meta.deserialize_skip``

        :param data dict: JSON-style object with instance data.
        :return: this instance
        """
        if not isinstance(data, dict):
            raise ValueError("Need to deserialize from a dict")

        try:
            skip = set(getattr(self._meta, "deserialize_skip", []))
        except AttributeError:  # _meta not available
            skip = []

        for key, value in list(data.items()):
            if key not in skip:
                value = self._deserialize_value(key, value)
                setattr(self, key, value)
        return self

    def _deserialize_value(self, key, value):
        if key.endswith("_at") and isinstance(value, str):
            value = make_date(value)
        return value

    def _serialize(self, skip_empty=True):
        """
        Serialise this instance into JSON-style request data.

        Filters out:
        * attribute names starting with ``_``
        * attribute values that are ``None`` (unless ``skip_empty`` is ``False``)
        * attribute values that are empty lists/tuples/dicts (unless ``skip_empty`` is ``False``)
        * attribute names in ``Meta.serialize_skip``
        * constants set on the model class

        Inner :py:class:`Model` instances get :py:meth:`._serialize` called on them.
        Date and datetime objects are converted into ISO 8601 strings.

        :param bool skip_empty: whether to skip attributes where the value is ``None``
        :rtype: dict
        """
        skip = set(getattr(self._meta, "serialize_skip", []))

        r = {}
        for k, v in list(self.__dict__.items()):
            if k.startswith("_"):
                continue
            elif k in skip:
                continue
            elif v is None and skip_empty:
                continue
            elif isinstance(v, (dict, list, tuple, set)) and len(v) == 0 and skip_empty:
                continue
            else:
                r[k] = self._serialize_value(v)
        return r

    def _serialize_value(self, value):
        """
        Called by :py:meth:`._serialize` to serialise an individual value.
        """
        if isinstance(value, (list, tuple, set)):
            return [self._serialize_value(v) for v in value]
        elif isinstance(value, dict):
            return dict([(k, self._serialize_value(v)) for k, v in list(value.items())])
        elif isinstance(value, ModelBase):
            return value._serialize()
        elif isinstance(value, datetime.date):  # includes datetime.datetime
            return value.isoformat()
        else:
            return value


class ModelBase(SerializableBase, metaclass=ModelMeta):
    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self)

    def __str__(self):
        s = str(getattr(self, "id", None))
        if getattr(self, "title", None):
            s += " - %s" % self.title
        return s

    def __init__(self, **kwargs):
        self._manager = None
        self.id = None
        for k, v in list(kwargs.items()):
            setattr(self, k, v)

    def __eq__(self, other):
        # is it a Model?
        if not isinstance(other, ModelBase):
            return False

        # is it a child or parent class?
        if not (
            issubclass(self.__class__, other.__class__)
            or issubclass(other.__class__, self.__class__)
        ):
            return False

        # am I bound?
        if not hasattr(self, "id"):
            return False

        # does it's id match mine?
        if getattr(other, "id", None) != self.id:
            return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def _is_bound(self):
        return bool(
            getattr(self, "id", None) is not None
            and getattr(self, "url", None)
            and getattr(self, "_manager", None)
        )

    @property
    def _client(self):
        if not self._manager:
            raise ValueError("%r must be bound to access a Client" % self)
        return self._manager.client

    def _deserialize(self, data, manager):
        """
        Deserialise from JSON response data.

        String items named ``*_at`` are turned into dates.
        Dict items named ``_by`` are turned into Users.

        :param data dict: JSON-style object with instance data.
        :return: this instance
        """
        if not issubclass(self.__class__, manager.model):
            raise TypeError(
                "Manager %s is for %s, expecting %s"
                % (
                    manager.__class__.__name__,
                    manager.model.__name__,
                    self.__class__.__name__,
                )
            )

        self._manager = manager
        return super(ModelBase, self)._deserialize(data)

    def _deserialize_value(self, key, value):
        from .users import User

        if key.endswith("_by") and isinstance(value, dict):
            value = User()._deserialize(value, self._manager.client.get_manager(User))
        else:
            value = super(ModelBase, self)._deserialize_value(key, value)
        return value


class Model(ModelBase):
    """
    Base class for models with managers.

    Model subclasses need a Meta class, which (in particular)
    links to their Manager class:

        class FooManager(Manager):
            ...

        class Foo(Model):
            class Meta:
                # Manager class
                manager = FooManager
                # Attributes available for ordering
                ordering_attributes = []
                # Attributes available for filtering
                filter_attributes = []
                # For the default implementation of ._serialize(), attributes to skip.
                serialize_skip = []
                # For the default implementation of ._deserialize(), attributes to skip.
                deserialize_skip = []
            ...
    """

    @is_bound
    def refresh(self):
        """
        Refresh this model from the server.

        Updates attributes with the server-defined values. This is useful where the Model
        instance came from a partial response (eg. a list query) and additional details
        are required.

        Existing attribute values will be overwritten.
        """
        r = self._client.request("GET", self.url)
        return self._deserialize(r.json(), self._manager)


class InnerModel(ModelBase):
    """
    Base class for Inner Models.

    These are models that are nested inside attributes of another model,
    and are saved and loaded as part of the parent model.
    """

    def __init__(self, **kwargs):
        self._parent = None
        super(InnerModel, self).__init__(**kwargs)

    @property
    def _is_bound(self):
        return bool(self._parent and self._parent._is_bound)

    @property
    def _client(self):
        return self._parent._client

    def _deserialize(self, data, manager, parent):
        self._parent = parent
        return super(InnerModel, self)._deserialize(data, manager)



# -*- coding: utf-8 -*-

"""
koordinates.publishing
======================

The `Group Publishing API <https://help.koordinates.com/api/publisher-admin-api/publishing-api/>`_
allows for draft versions of versioned objects Documents,
Layers, and Tables) to be scheduled for publishing together.
"""
import logging

from . import base
from .utils import is_bound


logger = logging.getLogger(__name__)


class PublishManager(base.Manager):
    """
    Accessor for querying Publish groups.

    Access via the ``publishing`` property of a :py:class:`koordinates.client.Client` instance.
    """

    _URL_KEY = "PUBLISH"

    def create(self, publish):
        """
        Creates a new publish group.
        """
        target_url = self.client.get_url("PUBLISH", "POST", "create")
        r = self.client.request("POST", target_url, json=publish._serialize())
        return self.create_from_result(r.json())


class Publish(base.Model):
    """
    Represents an active publishing group.
    """

    class Meta:
        manager = PublishManager
        attribute_filter_candidates = (
            "state",
            "reference",
        )

    PUBLISH_STRATEGY_TOGETHER = "together"
    PUBLISH_STRATEGY_INDIVIDUAL = "individual"

    ERROR_STRATEGY_ABORT = "abort"
    ERROR_STRATEGY_IGNORE = "ignore"

    def __init__(self, **kwargs):
        self.items = []
        super(Publish, self).__init__(**kwargs)

    @is_bound
    def cancel(self):
        """ Cancel a pending publish task """
        target_url = self._client.get_url(
            "PUBLISH", "DELETE", "single", {"id": self.id}
        )
        r = self._client.request("DELETE", target_url)
        logger.info("cancel(): %s", r.status_code)

    def get_items(self):
        """
        Return the item models associated with this Publish group.
        """
        from .layers import Layer

        # no expansion support, just URLs
        results = []
        for url in self.items:
            if "/layers/" in url:
                r = self._client.request("GET", url)
                results.append(
                    self._client.get_manager(Layer).create_from_result(r.json())
                )
            else:
                raise NotImplementedError("No support for %s" % url)
        return results

    def add_layer_item(self, layer):
        """
        Adds a Layer to the publish group.
        """
        if not layer.is_draft_version:
            raise ValueError("Layer isn't a draft version")

        self.items.append(layer.latest_version)

    def add_table_item(self, table):
        """
        Adds a Table to the publish group.
        """
        if not table.is_draft_version:
            raise ValueError("Table isn't a draft version")

        self.items.append(table.latest_version)



# -*- coding: utf-8 -*-

"""
koordinates.permissions
==================

For getting, editing and updating permissions via
the `Permissions API <https://help.koordinates.com/api/publisher-admin-api/permissions-api/>`_.

"""
import logging

from . import base
from . import exceptions
from .users import Group, User


logger = logging.getLogger(__name__)


class PermissionManager(base.InnerManager):
    """
    Accessor for querying and updating permissions.

    Access via the ``permissions`` property of :py:class:`koordinates.layers.Layer`
    or :py:class:`koordinates.sets.Set` instances.
    """

    _URL_KEY = "PERMISSION"

    def __init__(self, client, parent_object):
        super(PermissionManager, self).__init__(client, parent_object._manager)
        self.parent_object = parent_object

    def create(self, permission):
        """
        Create single permission for the given object.

        :param Permission permission: A single Permission object to be set.
        """
        parent_url = self.client.get_url(
            self.parent_object._manager._URL_KEY,
            "GET",
            "single",
            {"id": self.parent_object.id},
        )
        target_url = parent_url + self.client.get_url_path(
            self._URL_KEY, "POST", "single"
        )
        r = self.client.request("POST", target_url, json=permission._serialize())
        return permission._deserialize(r.json(), self)

    def set(self, permissions):
        """
        Set the object permissions. If the parent object already has permissions, they will be overwritten.

        :param [] permissions: A group of Permission objects to be set.
        """
        parent_url = self.client.get_url(
            self.parent_object._manager._URL_KEY,
            "GET",
            "single",
            {"id": self.parent_object.id},
        )
        target_url = parent_url + self.client.get_url_path(
            self._URL_KEY, "PUT", "multi"
        )
        r = self.client.request("PUT", target_url, json=permissions)
        if r.status_code != 201:
            raise exceptions.ServerError(
                "Expected 201 response, got %s: %s" % (r.status_code, target_url)
            )
        return self.list()

    def list(self):
        """
        List permissions for the given object.
        """
        parent_url = self.client.get_url(
            self.parent_object._manager._URL_KEY,
            "GET",
            "single",
            {"id": self.parent_object.id},
        )
        target_url = parent_url + self.client.get_url_path(
            self._URL_KEY, "GET", "multi"
        )
        return base.Query(self, target_url)

    def get(self, permission_id, expand=[]):
        """
        List a specific permisison for the given object.

        :param str permission_id: the id of the Permission to be listed.
        """
        parent_url = self.client.get_url(
            self.parent_object._manager._URL_KEY,
            "GET",
            "single",
            {"id": self.parent_object.id},
        )
        target_url = parent_url + self.client.get_url_path(
            self._URL_KEY, "GET", "single", {"permission_id": permission_id}
        )
        return self._get(target_url, expand=expand)


class Permission(base.InnerModel):
    """
    Represents a permissions for a specific :py:class:`koordinates.layers.Layer`
    or :py:class:`koordinates.sets.Set` instance.
    """

    def _deserialize(self, data, manager):
        super(Permission, self)._deserialize(data, manager, manager.parent_object)
        self.group = (
            Group()._deserialize(data["group"], manager.client.get_manager(Group))
            if data.get("group")
            else None
        )
        self.user = (
            User()._deserialize(data["user"], manager.client.get_manager(User))
            if data.get("user")
            else None
        )
        return self

    class Meta:
        manager = PermissionManager


class PermissionObjectMixin(object):
    """
    Mixin to be used in any Koordinates Object class that supports permissions. Also remember to set the Meta
    required fields:
    ```
        class Meta:
            serialize_skip = ('permissions',)
            deserialize_skip = ('permissions',)
    ```
    """

    _perms = None

    @property
    def permissions(self):
        if not self._perms:
            self._perms = PermissionManager(self._client, self)
        return self._perms



"""
koordinates.exports
===================

The `Exports API <https://support.koordinates.com/hc/en-us/articles/208580966>`_
provides an interface to create exports and download data from a Koordinates site.
"""
import collections
import contextlib
import logging
import os


from . import base
from . import exceptions
from .utils import is_bound


logger = logging.getLogger(__name__)


class CropLayerManager(base.Manager):
    """
    Accessor for querying Crop Layers.

    Access via the ``exports.croplayers`` property of a :py:class:`koordinates.client.Client` instance.
    """

    _URL_KEY = "CROPLAYER"

    def get_feature(self, croplayer_id, cropfeature_id):
        """
        Gets a crop feature

        :param int croplayer_id: ID of a cropping layer
        :param int cropfeature_id: ID of a cropping feature
        :rtype: CropFeature
        """
        target_url = self.client.get_url(
            "CROPFEATURE",
            "GET",
            "single",
            {"croplayer_id": croplayer_id, "cropfeature_id": cropfeature_id},
        )
        return self.client.get_manager(CropFeature)._get(target_url)


class CropLayer(base.Model):
    """
    A crop layer provides features that can be used to crop exports to a geographic extent.
    """

    class Meta:
        manager = CropLayerManager
        relations = {
            "features": ["CropFeature"],
        }

    @is_bound
    def get_feature(self, cropfeature_id):
        """
        Gets a crop feature

        :param int cropfeature_id: ID of a cropping feature
        :rtype: CropFeature
        """
        return self._manager.get_feature(self.id, cropfeature_id)


class CropFeatureManager(base.Manager):
    """
    Accessor for querying Crop Features.
    """

    _URL_KEY = "CROPFEATURE"


class CropFeature(base.Model):
    """
    A crop feature provides complex pre-defined geographic extents for cropping
    and clipping Exports.
    """

    class Meta:
        manager = CropFeatureManager
        relations = {
            "layer": CropLayer,
        }

    def _serialize(self):
        return self.url


class ExportManager(base.Manager):
    """
    Accessor for querying and creating Exports.

    Access via the ``exports`` property of a :py:class:`koordinates.client.Client` instance.
    """

    _URL_KEY = "EXPORT"
    _options_cache = None

    @property
    def croplayers(self):
        """
        Returns a manager for querying and listing CropLayer models

        :rtype: CropLayerManager
        """
        return self.client.get_manager(CropLayer)

    def create(self, export):
        """
        Create and start processing a new Export.

        :param Export export: The Export to create.
        :rtype: Export
        """
        target_url = self.client.get_url(self._URL_KEY, "POST", "create")
        r = self.client.request("POST", target_url, json=export._serialize())
        return export._deserialize(r.json(), self)

    def validate(self, export):
        """
        Validates an Export.


        :param Export export:
        :rtype: ExportValidationResponse
        """
        target_url = self.client.get_url(self._URL_KEY, "POST", "validate")
        response_object = ExportValidationResponse()
        r = self.client.request("POST", target_url, json=export._serialize())
        return response_object._deserialize(r.json())

    def cancel(self, export_id):
        target_url = self.client.get_url(
            self._URL_KEY, "DELETE", "single", {"id": export_id}
        )
        r = self.client.request("DELETE", target_url)
        return self.create_from_result(r.json())

    def _options(self):
        """
        Returns a raw options object

        :rtype: dict
        """
        if self._options_cache is None:
            target_url = self.client.get_url(self._URL_KEY, "OPTIONS", "options")
            r = self.client.request("OPTIONS", target_url)
            self._options_cache = r.json()
        return self._options_cache

    def get_formats(self):
        """
        Returns a dictionary of format options keyed by data kind.

        .. code-block:: python

            {
                "vector": {
                    "application/x-ogc-gpkg": "GeoPackage",
                    "application/x-zipped-shp": "Shapefile",
                    #...
                },
                "table": {
                    "text/csv": "CSV (text/csv)",
                    "application/x-ogc-gpkg": "GeoPackage",
                    #...
                },
                "raster": {
                    "image/jpeg": "JPEG",
                    "image/jp2": "JPEG2000",
                    #...
                },
                "grid": {
                    "application/x-ogc-aaigrid": "ASCII Grid",
                    "image/tiff;subtype=geotiff": "GeoTIFF",
                    #...
                },
                "rat": {
                    "application/x-erdas-hfa": "ERDAS Imagine",
                    #...
                }
            }

        :rtype: dict
        """
        format_opts = self._options()["actions"]["POST"]["formats"]["children"]
        r = {}
        for kind, kind_opts in list(format_opts.items()):
            r[kind] = {c["value"]: c["display_name"] for c in kind_opts["choices"]}
        return r


class ExportValidationResponse(base.SerializableBase):
    """
    Repsonse returned by Export validation requests.
    """

    def __init__(self, **kwargs):
        self.items = []
        super(ExportValidationResponse, self).__init__(**kwargs)

    def get_reasons(self):
        r = {}
        if self.invalid_reasons:
            r["__all__"] = self.invalid_reasons[:]
        for item in self.items:
            if item["invalid_reasons"]:
                r[item["item"]] = item["invalid_reasons"][:]
        return r


class DownloadError(exceptions.ClientError):
    pass


class Export(base.Model):
    """
    An export is a request to extract data from a Koordinates site into an archive for downloading

    :Example:

    >>> export = koordinates.Export()
    >>> export.crs = "EPSG:4326"
    >>> export.formats = {
            "vector": "application/x-zipped-shp"
        }
    >>> export.add_item(layer)
    >>> client.exports.create(export)
    """

    class Meta:
        manager = ExportManager

    def __init__(self, **kwargs):
        self.items = []
        super(Export, self).__init__(**kwargs)

    def add_item(self, item, **options):
        """
        Add a layer or table item to the export.

        :param Layer|Table item: The Layer or Table to add
        :rtype: self
        """
        export_item = {
            "item": item.url,
        }
        export_item.update(options)
        self.items.append(export_item)
        return self

    def set_formats(self, **kinds):
        if not hasattr(self, "formats"):
            self.formats = {}

        for kind, data_format in list(kinds.items()):
            if data_format:
                self.formats[kind] = data_format
            elif kind in self.formats:
                del self.formats[kind]

        return self.formats

    @is_bound
    def cancel(self):
        """
        Cancel the export processing
        """
        return self._manage.cancel(self.id)

    @is_bound
    def download(self, path, progress_callback=None, chunk_size=1024 ** 2):
        """
        Download the export archive.

        .. warning::

            If you pass this function an open file-like object as the ``path``
            parameter, the function will not close that file for you.

        If a ``path`` parameter is a directory, this function will use the
        Export name to determine the name of the file (returned). If the
        calculated download file path already exists, this function will raise
        a DownloadError.

        You can also specify the filename as a string. This will be passed to
        the built-in :func:`open` and we will read the content into the file.

        Instead, if you want to manage the file object yourself, you need to
        provide either a :class:`io.BytesIO` object or a file opened with the
        `'b'` flag. See the two examples below for more details.

        :param path: Either a string with the path to the location
            to save the response content, or a file-like object expecting bytes.
        :param function progress_callback: An optional callback
                function which receives upload progress notifications. The function should take two
                arguments: the number of bytes recieved, and the total number of bytes to recieve.
        :param int chunk_size: Chunk size in bytes for streaming large downloads and progress reporting. 1MB by default

        :returns The name of the automatic filename that would be used.
        :rtype: str
        """
        if not self.download_url or self.state != "complete":
            raise DownloadError("Download not available")

        # ignore parsing the Content-Disposition header, since we know the name
        download_filename = "{}.zip".format(self.name)
        fd = None

        if isinstance(getattr(path, "write", None), collections.Callable):
            # already open file-like object
            fd = path
        elif os.path.isdir(path):
            # directory to download to, using the export name
            path = os.path.join(path, download_filename)
            # do not allow overwriting
            if os.path.exists(path):
                raise DownloadError("Download file already exists: %s" % path)
        elif path:
            # fully qualified file path
            # allow overwriting
            pass
        elif not path:
            raise DownloadError("Empty download file path")

        with contextlib.ExitStack() as stack:
            if not fd:
                fd = open(path, "wb")
                # only close a file we open
                stack.callback(fd.close)

            r = self._manager.client.request("GET", self.download_url, stream=True)
            stack.callback(r.close)

            bytes_written = 0
            try:
                bytes_total = int(r.headers.get("content-length", None))
            except TypeError:
                bytes_total = None

            if progress_callback:
                # initial callback (0%)
                progress_callback(bytes_written, bytes_total)

            for chunk in r.iter_content(chunk_size=chunk_size):
                fd.write(chunk)
                bytes_written += len(chunk)
                if progress_callback:
                    progress_callback(bytes_written, bytes_total)

        return download_filename



# -*- coding: utf-8 -*-
"""
Koordinates Python API Client Library

:copyright: (c) Koordinates Limited.
:license: BSD, see LICENSE for more details.
"""

__version__ = "0.6.0"

from .exceptions import (
    KoordinatesException,
    ClientError,
    ClientValidationError,
    InvalidAPIVersion,
    ServerError,
    BadRequest,
    AuthenticationError,
    Forbidden,
    NotFound,
    NotAllowed,
    Conflict,
    RateLimitExceeded,
    InternalServerError,
    ServiceUnvailable,
)

from .client import Client
from .layers import Layer, Table
from .licenses import License
from .metadata import Metadata
from .publishing import Publish
from .sets import Set
from .sources import Source, UploadSource
from .users import Group, User
from .permissions import Permission
from .exports import Export, CropLayer, DownloadError



# -*- coding: utf-8 -*-

"""
koordinates.licenses
===================

The `License API <https://help.koordinates.com/api/publisher-admin-api/license-api/>`_
provides access to licenses that can be applied to layers, tables, sources
and documents. It provides access to Creative Commons API by property
lookup, as well as custom and built-in licenses.
"""

from . import base
from . import exceptions


class LicenseManager(base.Manager):
    """
    Accessor for querying licenses.

    Access via the ``licenses`` property of a :py:class:`koordinates.client.Client` instance.
    """

    _URL_KEY = "LICENSE"

    def get_creative_commons(self, slug, jurisdiction=None):
        """Returns the Creative Commons license for the given attributes.

        :param str slug: the type of Creative Commons license. It must start with
            ``cc-by`` and can optionally contain ``nc`` (non-commercial),
            ``sa`` (share-alike), ``nd`` (no derivatives) terms, seperated by
            hyphens. Note that a CC license cannot be both ``sa`` and ``nd``
        :param str jurisdiction: The jurisdiction for a ported Creative Commons
            license (eg. ``nz``), or ``None`` for unported/international licenses.
        :rtype: License
        """
        if not slug.startswith("cc-by"):
            raise exceptions.ClientValidationError("slug needs to start with 'cc-by'")

        if jurisdiction is None:
            jurisdiction = ""

        target_url = self.client.get_url(
            self._URL_KEY, "GET", "cc", {"slug": slug, "jurisdiction": jurisdiction}
        )
        return self._get(target_url)


class License(base.Model):
    """
    Represents a license that can be applied to layers, tables, sources, and documents.
    """

    class Meta:
        manager = LicenseManager



import logging

from . import base


logger = logging.getLogger(__name__)


class UserManager(base.Manager):
    _URL_KEY = "USER"


class User(base.Model):
    """ Represents a Koordinates User """

    class Meta:
        manager = UserManager


class GroupManager(base.Manager):
    _URL_KEY = "GROUP"


class Group(base.Model):
    """ Represents a Koordinates Group """

    class Meta:
        manager = GroupManager



class KoordinatesException(Exception):
    """ Base class for all koordinates module errors """

    def __init__(self, message, **kwargs):
        super(KoordinatesException, self).__init__(message)
        for k, v in list(kwargs.items()):
            setattr(self, k, v)


# Client Errors
class ClientError(KoordinatesException):
    """ Base class for client errors """

    pass


class ClientValidationError(ClientError):
    """ Client-side validation error """

    pass


class InvalidAPIVersion(ClientError):
    """ Invalid API Version """

    pass


# Server Errors
class ServerError(KoordinatesException):
    """ Base class for errors returned from the API Servers """

    @classmethod
    def from_requests_error(cls, err):
        """
        Raises a subclass of ServerError based on the HTTP response code.
        """
        import requests

        if isinstance(err, requests.HTTPError):
            status_code = err.response.status_code
            return HTTP_ERRORS.get(status_code, cls)(error=err, response=err.response)
        else:
            return cls(error=err)

    def __init__(self, message=None, error=None, response=None):
        self.response = response
        if not message:
            message = self._get_message(error, response)
        super(ServerError, self).__init__(message, error=error, response=response)

    def _get_message(self, error, response):
        if response:
            message = "%s %s" % (response.status_code, response.reason)
            try:
                # most API errors are of the form {"error": "description"}
                message += ": %s" % response.json()["error"]
            except:
                pass
        else:
            message = str(error)
        return message

    def __str__(self):
        if self.response:
            return "%s: %s" % (self.args[0], self.response.text)
        else:
            return super(ServerError, self).__str__()

    def __repr__(self):
        return "%s('%s')" % (self.__class__.__name__, str(self))


class BadRequest(ServerError):
    """ Invalid request data or parameters. Check your request. (400) """

    def _get_message(self, error, response):
        try:
            messages = []
            for field, errors in sorted(response.json().items()):
                errors = errors if isinstance(errors, str) else "; ".join(errors)
                messages.append("%s: %s" % (field, errors))
            return "\n".join(messages)
        except Exception as e:
            return super(BadRequest, self)._get_message(error, response)


class AuthenticationError(ServerError):
    """ The API token is invalid or expired. (401) """

    pass


class Forbidden(ServerError):
    """ The API token or user doesn't have access to perform the request. (403) """

    pass


class NotFound(ServerError):
    """ The requested object was not found. (404) """

    pass


class NotAllowed(ServerError):
    """ The requested action isn't available for this object. (405) """

    pass


class Conflict(ServerError):
    """ The requested action isn't available for this object due to a conflict. (409) """

    pass


class RateLimitExceeded(ServerError):
    """ The request has exceeded the API rate limit. Retry the request again later. (429) """

    pass


class InternalServerError(ServerError):
    """ An internal server error has occurred. (500) """

    pass


class ServiceUnvailable(ServerError):
    """ The Koordinates service is currently unavailable. (502/503/504) """

    pass


HTTP_ERRORS = {
    400: BadRequest,
    401: AuthenticationError,
    403: Forbidden,
    404: NotFound,
    405: NotAllowed,
    409: Conflict,
    422: BadRequest,
    429: RateLimitExceeded,
    500: InternalServerError,
    502: ServiceUnvailable,
    503: ServiceUnvailable,
    504: ServiceUnvailable,
}



# -*- coding: utf-8 -*-

"""
koordinates.sets
================

The `Sets API <https://help.koordinates.com/api/publisher-admin-api/sets-api/>`_
is used for grouping layers, tables and documents together.
"""

import logging

from koordinates.permissions import PermissionObjectMixin
from koordinates.users import Group
from koordinates.metadata import Metadata, MetadataManager
from koordinates import base
from koordinates.utils import is_bound
from .publishing import Publish

logger = logging.getLogger(__name__)


class SetManager(base.Manager):
    """
    Accessor for querying Sets.

    Access via the ``sets`` property of a :py:class:`koordinates.client.Client` instance.
    """

    _URL_KEY = "SET"

    def __init__(self, client):
        super(SetManager, self).__init__(client)
        # Inner model managers
        self.versions = SetVersionManager(client, self)
        self._data = SetDataManager(client, self)
        self._metadata = MetadataManager(self, client)

    def list_drafts(self):
        """
        A filterable list view of sets, returning the draft version of each set.
        If the most recent version has been published already, it won’t be returned here.
        """
        target_url = self.client.get_url("SET", "GET", "multidraft")
        return base.Query(self, target_url)

    def create(self, set):
        """
        Creates a new Set.
        """
        target_url = self.client.get_url("SET", "POST", "create")
        r = self.client.request("POST", target_url, json=set._serialize())
        return set._deserialize(r.json(), self)

    def list_versions(self, set_id):
        """
        Filterable list of versions of a set, always ordered newest to oldest.

        If the version’s source supports revisions, you can get a specific revision using
        ``.filter(data__source_revision=value)``. Specific values depend on the source type.
        Use ``data__source_revision__lt`` or ``data__source_revision__gte`` to filter
        using ``<`` or ``>=`` operators respectively.
        """
        target_url = self.client.get_url("SET_VERSION", "GET", "multi", {"id": set_id})
        return base.Query(
            self,
            target_url,
            valid_filter_attributes=("data",),
            valid_sort_attributes=(),
        )

    def get_version(self, set_id, version_id, expand=[]):
        """
        Get a specific version of a set.
        """
        target_url = self.client.get_url(
            "SET_VERSION", "GET", "single", {"id": set_id, "version_id": version_id},
        )
        return self._get(target_url, expand=expand)

    def get_draft(self, set_id, expand=[]):
        """
        Get the current draft version of a set.
        :raises NotFound: if there is no draft version.
        """
        target_url = self.client.get_url("SET_VERSION", "GET", "draft", {"id": set_id})
        return self._get(target_url, expand=expand)

    def get_published(self, set_id, expand=[]):
        """
        Get the latest published version of this set.
        :raises NotFound: if there is no published version.
        """
        target_url = self.client.get_url(
            "SET_VERSION", "GET", "published", {"id": set_id}
        )
        return self._get(target_url, expand=expand)

    def create_draft(self, set_id):
        """
        Creates a new draft version.

        :rtype: Client
        :return: the new version
        :raises 409 Conflict: if there is already a draft version for this set.
        """
        target_url = self.client.get_url(
            "SET_VERSION", "POST", "create", {"id": set_id}
        )
        r = self.client.request("POST", target_url, json={})
        return self.create_from_result(r.json())

    def set_metadata(self, set_id, fp):
        """
        Set the XML metadata on a set.

        :param file fp: file-like object to read the XML metadata from.
        """
        base_url = self.client.get_url("SET", "GET", "single", {"id": set_id})
        self._metadata.set(base_url, fp)


class Set(base.Model, PermissionObjectMixin):
    """
    Represents a single set grouping of layers, tables, and documents.
    """

    class Meta:
        manager = SetManager
        serialize_skip = ("permissions",)
        deserialize_skip = ("permissions",)

    def _deserialize(self, data, manager):
        super(Set, self)._deserialize(data, manager)
        self.group = (
            Group()._deserialize(data["group"], manager.client.get_manager(Group))
            if data.get("group")
            else None
        )
        self.metadata = (
            Metadata()._deserialize(data["metadata"], manager._metadata, self)
            if data.get("metadata")
            else None
        )
        self.version = (
            SetVersion()._deserialize(data["version"], manager.versions, self)
            if data.get("version")
            else None
        )
        return self

    @is_bound
    def set_metadata(self, fp, version_id=None):
        """
        Set the XML metadata on this draft version.

        :param file fp: file-like object to read the XML metadata from.
        :raises NotAllowed: if this version is already published.
        """
        if not version_id:
            version_id = self.version.id

        base_url = self._client.get_url(
            "SET_VERSION", "GET", "single", {"id": self.id, "version_id": version_id},
        )
        self._manager._metadata.set(base_url, fp)

        # reload myself
        r = self._client.request("GET", base_url)
        return self._deserialize(r.json(), self._manager)

    @property
    def is_published_version(self):
        """ Return if this version is the published version of a layer """
        pub_ver = getattr(self, "published_version", None)
        this_ver = getattr(self, "this_version", None)
        return this_ver and pub_ver and (this_ver == pub_ver)

    @property
    def is_draft_version(self):
        """ Return if this version is the draft version of a layer """
        pub_ver = getattr(self, "published_version", None)
        latest_ver = getattr(self, "latest_version", None)
        this_ver = getattr(self, "this_version", None)
        return (
            this_ver
            and latest_ver
            and (this_ver == latest_ver)
            and (latest_ver != pub_ver)
        )

    @is_bound
    def list_versions(self):
        """
        Filterable list of versions of a set, always ordered newest to oldest.

        If the version’s source supports revisions, you can get a specific revision using
        ``.filter(data__source_revision=value)``. Specific values depend on the source type.
        Use ``data__source_revision__lt`` or ``data__source_revision__gte`` to filter
        using ``<`` or ``>=`` operators respectively.
        """
        target_url = self._client.get_url(
            "SET_VERSION", "GET", "multi", {"id": self.id}
        )
        return base.Query(
            self._manager,
            target_url,
            valid_filter_attributes=("data",),
            valid_sort_attributes=(),
        )

    @is_bound
    def get_version(self, version_id, expand=()):
        """
        Get a specific version of this set
        """
        target_url = self._client.get_url(
            "SET_VERSION", "GET", "single", {"id": self.id, "version_id": version_id},
        )
        return self._manager._get(target_url, expand=expand)

    @is_bound
    def get_draft_version(self, expand=()):
        """
        Get the current draft version of this set.
        :raises NotFound: if there is no draft version.
        """
        target_url = self._client.get_url(
            "SET_VERSION", "GET", "draft", {"id": self.id}
        )
        return self._manager._get(target_url, expand=expand)

    @is_bound
    def get_published_version(self, expand=()):
        """
        Get the latest published version of this set.
        :raises NotFound: if there is no published version.
        """
        target_url = self._client.get_url(
            "SET_VERSION", "GET", "published", {"id": self.id}
        )
        return self._manager._get(target_url, expand=expand)

    @is_bound
    def publish(self, version_id=None):
        """
        Creates a publish task for this version.

        :return: the publish task
        :rtype: Publish
        :raises Conflict: If the version is already published, or already has a publish job.
        """
        if not version_id:
            version_id = self.version.id

        target_url = self._client.get_url(
            "SET_VERSION", "POST", "publish", {"id": self.id, "version_id": version_id},
        )
        r = self._client.request("POST", target_url, json={})
        return self._client.get_manager(Publish).create_from_result(r.json())

    @is_bound
    def save(self):
        """
        Edits this draft version.

        :raises NotAllowed: if the version is already published.
        """
        target_url = self._client.get_url(
            "SET_VERSION",
            "PUT",
            "edit",
            {"id": self.id, "version_id": self.version.id},
        )
        r = self._client.request("PUT", target_url, json=self._serialize())
        return self._deserialize(r.json(), self._manager)

    @is_bound
    def delete_version(self, version_id=None):
        """
        Deletes this draft version (revert to published)

        :raises NotAllowed: if this version is already published.
        :raises Conflict: if this version is already deleted.
        """
        if not version_id:
            version_id = self.version.id

        target_url = self._client.get_url(
            "SET_VERSION",
            "DELETE",
            "single",
            {"id": self.id, "version_id": version_id},
        )
        self._client.request("DELETE", target_url)


class SetVersionManager(base.InnerManager):
    _URL_KEY = "SET_VERSION"


class SetVersion(base.InnerModel):
    """
    Represents the ``version`` property of a :py:class:`koordinates.client.Client` instance.
    """

    class Meta:
        manager = SetVersionManager


class SetDataManager(base.InnerManager):
    _URL_KEY = "DATA"


class SetData(base.InnerModel):
    """
    Represents the ``data`` property of a :py:class:`koordinates.client.Client` instance.
    """

    class Meta:
        manager = SetDataManager



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import re
from codecs import open


version = ""
with open("koordinates/__init__.py", "r") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)
with open("README.md", "r", "utf-8") as f:
    readme = f.read()

setup(
    name="koordinates",
    packages=[
        "koordinates",
    ],
    version=version,
    description="A Python client library for a number of Koordinates web APIs",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Koordinates Limited",
    author_email="support@koordinates.com",
    url="https://github.com/koordinates/python-client",
    keywords="koordinates api",
    license="BSD",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: GIS",
    ],
    python_requires=">=3.5",
    install_requires=[
        "python-dateutil>=2,<3",
        "pytz",
        "requests>=2.5,<3",
        "requests-toolbelt>=0.4,<1",
    ],
    tests_require=[
        "pytest>=3.3",
        "responses>=0.3",
        "coverage>=3.7,<4",
    ],
    zip_safe=False,
)



# -*- coding: utf-8 -*-

"""
koordinates.sources
===================

The `Data Sources API <https://help.koordinates.com/api/publisher-admin-api/data-sources-api/>`_
provides read+write access to sources, datasources and scans.

A source points to a place where Koordinates can get data from. Sources can contain any number of datasources.

A datasource is a single dataset from a source. One or more datasources may be imported to create a layer,
table or document.

A scan examines a source to find out about what datasources the source provides. Until a source is scanned
it may appear to have no datasources.
"""
import collections
import json
import logging
import mimetypes
import os

from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

from . import base
from .exceptions import ClientValidationError
from .metadata import Metadata, MetadataManager
from .permissions import PermissionObjectMixin
from .users import Group, User
from .utils import is_bound


logger = logging.getLogger(__name__)


class SourceManager(base.Manager):
    """
    Accessor for querying Sources.

    Access via the ``sources`` property of a :py:class:`koordinates.client.Client` instance.
    """

    _URL_KEY = "SOURCE"

    def __init__(self, client):
        super(SourceManager, self).__init__(client)
        # Inner model managers
        self._metadata = MetadataManager(client, self)

    def create(self, source, upload_progress_callback=None):
        """
        Creates a new source.

        :param str source: The populated Source object to create.
        :param function upload_progress_callback: For an UploadSource object, an optional callback
                    function which receives upload progress notifications. The function should take two
                    arguments: the number of bytes sent, and the total number of bytes to send.
        :rtype: Source
        """
        # Delegate to the class itself
        return source._create(self, callback=upload_progress_callback)

    def list_datasources(self, source_id):
        """
        Filterable list of Datasources for a Source.
        """
        target_url = self.client.get_url(
            "DATASOURCE", "GET", "multi", {"source_id": source_id}
        )
        return base.Query(self.client.get_manager(Datasource), target_url)

    def get_datasource(self, source_id, datasource_id):
        """
        Get a Datasource object

        :rtype: Datasource
        """
        target_url = self.client.get_url(
            "DATASOURCE",
            "GET",
            "single",
            {"source_id": source_id, "datasource_id": datasource_id},
        )
        return self.client.get_manager(Datasource)._get(target_url)

    def list_scans(self, source_id=None):
        """
        Filterable list of Scans for a Source.
        Ordered newest to oldest by default
        """
        if source_id:
            target_url = self.client.get_url(
                "SCAN", "GET", "multi", {"source_id": source_id}
            )
        else:
            target_url = self.client.get_ulr("SCAN", "GET", "all")
        return base.Query(self.client.get_manager(Scan), target_url)

    def get_scan(self, source_id, scan_id):
        """
        Get a Scan object

        :rtype: Scan
        """
        target_url = self.client.get_url(
            "SCAN", "GET", "single", {"source_id": source_id, "scan_id": scan_id}
        )
        return self.client.get_manager(Scan)._get(target_url)

    def get_scan_log_lines(self, source_id, scan_id):
        """
        Get the log text for a Scan

        :rtype: Iterator over log lines.
        """
        return self.client.get_manager(Scan).get_log_lines(
            source_id=source_id, scan_id=scan_id
        )

    def start_scan(self, source_id):
        """
        Start a new scan of a Source.

        :rtype: Scan
        """
        target_url = self.client.get_url(
            "SCAN", "POST", "create", {"source_id": source_id}
        )
        r = self.client.request("POST", target_url, json={})
        return self.client.get_manager(Scan).create_from_result(r.json())


class Source(base.Model, PermissionObjectMixin):
    """
    A source points to a place where Koordinates can get data from. Sources can contain any number
    of datasources.
    """

    class Meta:
        manager = SourceManager
        relations = {
            "scans": ["Scan"],
            "datasources": ["Datasource"],
        }
        serialize_skip = ("permissions",)
        deserialize_skip = ("permissions",)

    TYPE_INFO = "info"
    TYPE_UPLOAD = "upload"
    TYPE_CIFS = "cifs"
    TYPE_ARCGIS = "arcgis"
    TYPE_WFS = "wfs"
    TYPE_POSTGRES = "postgres"

    def __init__(self, **kwargs):
        self.items = []
        super(Source, self).__init__(**kwargs)

    def _deserialize(self, data, manager):
        super(Source, self)._deserialize(data, manager)
        self.group = (
            Group()._deserialize(data["group"], manager.client.get_manager(Group))
            if data.get("group")
            else None
        )
        self.user = (
            User()._deserialize(data["user"], manager.client.get_manager(User))
            if data.get("user")
            else None
        )
        self.metadata = (
            Metadata()._deserialize(data["metadata"], manager._metadata, self)
            if data.get("metadata")
            else None
        )
        return self

    def _create(self, manager, *args, **kwargs):
        target_url = manager.client.get_url("SOURCE", "POST", "create")
        r = manager.client.request("POST", target_url, json=self._serialize())
        return manager.create_from_result(r.json())

    @is_bound
    def save(self, with_data=False):
        """
        Edits this Source
        """
        r = self._client.request(
            "PUT", self.url, json=self._serialize(with_data=with_data)
        )
        return self._deserialize(r.json(), self._manager)

    @is_bound
    def delete(self):
        """ Delete this source """
        r = self._client.request("DELETE", self.url)
        logger.info("delete(): %s", r.status_code)

    @is_bound
    def get_datasource(self, datasource_id):
        return self._manager.get_datasource(self.id, datasource_id)

    @is_bound
    def list_datasources(self):
        return self._manager.list_datasources(self.id)

    @is_bound
    def list_scans(self):
        return self._manager.list_scans(self.id)

    @is_bound
    def get_latest_scan(self):
        results = self._manager.list_scans(self.id).expand()[:1]
        if results:
            return results[0]

    @is_bound
    def get_scan(self, scan_id):
        return self._manager.get_scan(self.id, scan_id)


class UploadSource(Source):
    """
    Subclass of Source used for uploads of files and archives, which are automatically scanned.

    :Example:

    >>> upload = koordinates.UploadSource()
    >>> upload.user = 5
    >>> upload.title = "upload_source example"
    >>> upload.add_file('/path/to/data.zip')
    >>> upload = client.sources.create(upload)

    """

    class Meta:
        manager = SourceManager

    def __init__(self, *args, **kwargs):
        super(UploadSource, self).__init__(*args, **kwargs)
        self.type = self.TYPE_UPLOAD
        self._files = collections.OrderedDict()

    def _create(self, manager, callback=None):
        if self.type != self.TYPE_UPLOAD:
            raise ClientValidationError("Model/type mismatch")

        fields = {
            "source": (None, json.dumps(self._serialize()), "application/json"),
        }

        def wrapped_callback(monitor):
            callback(monitor.bytes_read, monitor.len)

        opened_files = []
        try:
            for i, (upload_path, (file_or_path, content_type)) in enumerate(
                self._files.items()
            ):
                if isinstance(file_or_path, str):
                    # path
                    fp = open(file_or_path, "rb")
                    opened_files.append(fp)
                else:
                    # file-like object
                    fp = file_or_path

                field = [
                    upload_path,
                    fp,
                ]

                if content_type:
                    field.append(content_type)

                fields["file%d" % i] = tuple(field)

            target_url = manager.client.get_url("SOURCE", "POST", "create")
            e = MultipartEncoder(fields=fields)
            m = MultipartEncoderMonitor(e, wrapped_callback if callback else None)

            r = manager.client.request(
                "POST", target_url, data=m, headers={"Content-Type": m.content_type}
            )
        finally:
            for fp in opened_files:
                fp.close()

        return manager.create_from_result(r.json())

    def add_file(self, fp, upload_path=None, content_type=None):
        """
        Add a single file or archive to upload.

        To add metadata records with a file, add a .xml file with the same upload path basename
        eg. ``points-with-metadata.geojson`` & ``points-with-metadata.xml``
        Datasource XML must be in one of these three formats:

          - ISO 19115/19139
          - FGDC CSDGM
          - Dublin Core (OAI-PMH)

        :param fp: File to upload into this source, can be a path or a file-like object.
        :type fp: str or file
        :param str upload_path: relative path to store the file as within the source (eg. ``folder/0001.tif``). \
                                By default it will use ``fp``, either the filename from a path or the ``.name`` \
                                attribute of a file-like object.
        :param str content_type: Content-Type of the file. By default it will attempt to auto-detect from the \
                                 file/upload_path.
        """
        if isinstance(fp, str):
            # path
            if not os.path.isfile(fp):
                raise ClientValidationError("Invalid file: %s", fp)
            if not upload_path:
                upload_path = os.path.split(fp)[1]
        else:
            # file-like object
            if not upload_path:
                upload_path = os.path.split(fp.name)[1]

        content_type = (
            content_type or mimetypes.guess_type(upload_path, strict=False)[0]
        )
        if upload_path in self._files:
            raise ClientValidationError("Duplicate upload path: %s" % upload_path)

        self._files[upload_path] = (fp, content_type)
        logger.debug(
            "UploadSource.add_file: %s -> %s (%s)", repr(fp), upload_path, content_type
        )


class ScanManager(base.Manager):
    _URL_KEY = "SCAN"

    def get_log_lines(self, source_id, scan_id):
        """
        Get the log text for a scan object
        :rtype: Iterator over log lines.
        """
        target_url = self.client.get_url(
            "SCAN", "GET", "log", {"source_id": source_id, "scan_id": scan_id}
        )
        r = self.client.request(
            "GET", target_url, headers={"Accept": "text/plain"}, stream=True
        )
        return r.iter_lines(decode_unicode=True)


class Scan(base.Model):
    """
    A scan operation examines a source to find out about what datasources the source provides.
    """

    class Meta:
        manager = ScanManager
        relations = {
            "source": Source,
        }
        filter_attributes = (
            "status",
            "started_at",
            "completed_at",
        )
        ordering_attributes = (
            "id",
            "started_at",
            "completed_at",
        )

    @is_bound
    def cancel(self):
        """
        Cancel a running Scan.
        """
        r = self._client.request("DELETE", self.url)
        logger.info("Scan:cancel(): %s", r.status_code)

    @is_bound
    def get_log_lines(self):
        """
        Get the log text for a scan object

        :rtype: Iterator over log lines.
        """
        rel = self._client.reverse_url("SCAN", self.url)
        return self._manager.get_log_lines(**rel)


class DatasourceManager(base.Manager):
    _URL_KEY = "DATASOURCE"

    def __init__(self, client):
        super(DatasourceManager, self).__init__(client)
        # Inner model managers
        self._metadata = MetadataManager(client, self)


class Datasource(base.Model):
    """
    A datasource is a single dataset from a source. One or more datasources may be imported to create a layer,
    table or document.
    """

    class Meta:
        manager = DatasourceManager
        relations = {
            "source": Source,
        }

    def _deserialize(self, data, manager):
        super(Datasource, self)._deserialize(data, manager)
        self.metadata = (
            Metadata()._deserialize(data["metadata"], manager._metadata, self)
            if data.get("metadata")
            else None
        )
        return self



# -*- coding: utf-8 -*-

"""
koordinates.utils
~~~~~~~~~~~~~~

This module provides utility functions and classes used in the Koordinates
Client Library

"""
import functools

import dateutil.parser


def is_bound(method):
    """
    Decorator that asserts the model instance is bound.

    Requires:
    1. an ``id`` attribute
    2. a ``url`` attribute
    2. a manager set
    """

    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        if not self._is_bound:
            raise ValueError("%r must be bound to call %s()" % (self, method.__name__))
        return method(self, *args, **kwargs)

    return wrapper


def make_date(v):
    """Returns a `DateTime` object, if `v` is a populated string, or an empty string.

    :param v: either a string parseable as a date/time; an empty string; or None
    :return either a `DateTime` corresponding to `v` or an empty String

    """
    if v == "" or v is None:
        return ""
    else:
        return dateutil.parser.parse(v)
